/*
 * bulk.c
 *
 *  Created on: Oct 03, 2009
 *      Author: jb
 */
#define DEBUG 0

#include <stdio.h>
#include <complex.h>
#include <math.h>
#include "orthodisp.h"

extern const double pi;

/* compute wave velocities cp and polarisation vectors pol for a wave
 * propagating in n-direction through bulk orthotropic material mat */
void get_ortho_bulk_wave(material mat, double n[3], double *cp, double *pol) {

	int i, j, imax1, imax2;
	double lambdar[3][3], lpol, npol, max1, max2, pol0[3][3], cp0[3];
	complex double lambdac[9], cpc[3], polc[9], dummy[9];
	complex double delta[9] = { 1., 0., 0., 0., 1., 0., 0., 0., 1. };

	/* get christoffel tensor [ 10^6 m^2 / s^2 ] */
	get_ortho_christoffel_tensor(mat, n, lambdar);
	
	/* it isnt a complex evp, but use existing complex lapack function anyway */
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			lambdac[i*3+j] = lambdar[i][j];
		}		
	}
	
	/* solve eigenvalue problem for wave velocities and polarisation vectors */
	clapack_complex_eigensystem(lambdac, delta, 3, cpc, dummy, polc);
	
	/* copy result to real vector / matrix */
	for (i=0; i<3; i++) {

		/* convert cp to m / s */
		cp0[i] = sqrt(creal(cpc[i])) * 1000;

		/* length of real part of polarisation vector and scalar product with
		 * propagation vector */
		lpol = 0;
		npol = 0;
		for (j=0; j<3; j++) {
			lpol += creal(polc[i*3+j]) * creal(polc[i*3+j]);
			npol += creal(polc[i*3+j]) * n[j];
		}
		lpol = sqrt(lpol) * sign(npol);

		/* normalize real part of eigenvectors */
		for (j=0; j<3; j++) {
			pol0[i][j] = creal(polc[i*3+j]) / lpol;
		}
	}

	/* sort by size of cp */
	max1 = -1;
	for (i=0; i<3; i++) {
		if (cp0[i] > max1) {
			max1 = cp0[i];
			imax1 = i;
		}
	}
	max2 = -1;
	for (i=0; i<3; i++) {
		if ((i != imax1) && (cp0[i] > max2)) {
			max2 = cp0[i];
			imax2 = i;
		}
	}

	/* rearrange */
	cp[0] = cp0[imax1];
	cp[1] = cp0[imax2];
	cp[2] = cp0[3-imax1-imax2];
	for (i=0; i<3; i++) {
		pol[i]     = pol0[imax1][i];
		pol[3 + i] = pol0[imax2][i];
		pol[6 + i] = pol0[3-imax1-imax2][i];
	}
	
}

void print_ortho_bulk_wave(char *fname, material mat) {

	int i, j;
	double angle, nbulk[3], cpbulk[3], polbulk[3][3];
	FILE *fbulk;

	fbulk = fopen(fname,"w");

	/* print header */
	fprintf(fbulk,"angle / rad\t   n0      \t   n1      \t   n2      \t");
	fprintf(fbulk,"\t  cp0 / m/s\t  cp1 / m/s\t  cp2 / m/s");
	for (i=0; i<3; i++) {
		fprintf(fbulk,"\t");
		for (j=0; j<3; j++) {
			fprintf(fbulk,"\t   p%d %d    ",i,j);
		}
	}
	fprintf(fbulk,"\n");

	/* loop over in-plane angle */
	for (angle = 0; angle < 2.*pi; angle += pi/24.) {
		
		/* set propagation direction */
		nbulk[0] = cos(angle);
		nbulk[1] = sin(angle);
		nbulk[2] = 0;
		
		/* get bulk wave velocities and polarisation */
		get_ortho_bulk_wave(mat, nbulk, cpbulk, (double *)polbulk);
		
		/* print */
		fprintf(fbulk,"%11.3e\t%11.3e\t%11.3e\t%11.3e\t",
				angle,nbulk[0],nbulk[1],nbulk[2]);
		for (i=0; i<3; i++) {
			fprintf(fbulk,"\t%11.3e",cpbulk[i]);
		}
		for (i=0; i<3; i++) {
			fprintf(fbulk,"\t");
			for (j=0; j<3; j++) {
				fprintf(fbulk,"\t%11.3e",polbulk[i][j]);
			}
		}
		fprintf(fbulk,"\n");
	}
	
	/* done */
	fclose(fbulk);
}
