/*
 * christoffel.c
 *
 *  Created on: Sep 2, 2009
 *      Author: jb
 */
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <math.h>
#include "orthodisp.h"

extern FILE *logfile;

/*
 * From Nayfeh '90 paper
 **********************************************************************/
void get_polarisation (
		double c[6][6],
		double rhocp2,
		double sintheta,
		complex double lz[6],
		complex double P[3][6],
		complex double D[3][6]) {

	complex double K[3][3], W[3][3];
//	complex double Pw[3];
	complex double lz2, delta = 0, sqrtW;
	double sin2 = sintheta * sintheta, maxW;
	int i, j0, j1, j2, jp, k0, k1, k2;

	for (i=0; i<6; i++) {
		lz2  = lz[i] * lz[i];

		/* this is the christoffel acoustic tensor for a
		 * propagation direction of [ 1  0  lz ] */
		K[0][0] = -rhocp2 + lz2 * c[4][4] + c[0][0] * sin2;
		K[1][1] = -rhocp2 + lz2 * c[3][3] + c[5][5] * sin2;
		K[2][2] = -rhocp2 + lz2 * c[2][2] + c[4][4] * sin2;

		K[0][1] = K[1][0] = lz2 * c[3][4] + c[0][5] * sin2;
		K[0][2] = K[2][0] = lz[i] * (c[0][2] + c[4][4]) * sintheta;
		K[1][2] = K[2][1] = lz[i] * (c[2][5] + c[3][4]) * sintheta;

		debugmsg(DEBUG,logfile,"polarisation K, lz = %11.3e%+10.3ei\n",creal(lz[i]),cimag(lz[i]));
		if (DEBUG) {
			print_c_matrix(logfile,(complex double *)K,3,3);
		}

		/* from wang, rokhlin 2002: first determine W tensor */
		for (j0 = 0; j0 < 3; j0++) {
			j1 = (j0 + 1) % 3;
			j2 = (j1 + 1) % 3;
			for (k0 = 0; k0 < 3; k0++) {
				k1 = (k0 + 1) % 3;
				k2 = (k1 + 1) % 3;
				W[j0][k0] = K[j1][k1] * K[j2][k2] - K[j1][k2] * K[j2][k1];
			}
			delta += W[j0][j0];
		}

		/* P[i] = sqrt(W[i][i]/delta)
		 * for correct sign, select first P component by choosing maximum absolute value
		 * sign of this component is arbitrary */
		maxW = -1;
		for (j0 = 0; j0 < 3; j0++) {
			sqrtW = csqrt(W[j0][j0]);
			if (cabs(sqrtW) > maxW) {
				maxW = cabs(sqrtW);
				jp = j0;
			}
		}
		P[jp][i] = csqrt(W[jp][jp]/delta);

		/* the other components are computed with correct sign from relation below */
		j1 = (jp + 1) % 3;
		j2 = (jp + 2) % 3;
		P[j1][i] = W[jp][j1] / (P[jp][i] * delta);
		P[j2][i] = W[jp][j2] / (P[jp][i] * delta);

		/* P is scaled so that x-component is 1.0 */
		for (j0 = 2; j0 >= 0; j0--) {
			P[j0][i] = P[j0][i] / P[0][i];
		}

		debugmsg(DEBUG,logfile,"lz = %11.3e%+10.3ei polarisation:\n",creal(lz[i]),cimag(lz[i]));
		debugmsg(DEBUG,logfile,"  U %11.3e%+10.3ei",creal(P[0][i]),cimag(P[0][i]));
		debugmsg(DEBUG,logfile,"  V %11.3e%+10.3ei",creal(P[1][i]),cimag(P[1][i]));
		debugmsg(DEBUG,logfile,"  W %11.3e%+10.3ei\n",creal(P[2][i]),cimag(P[2][i]));
	}

	for (i=0; i<6; i++) {
		D[0][i] = c[0][2] * sintheta * P[0][i] + c[2][5] * sintheta * P[1][i] + c[2][2] * lz[i] * P[2][i];  // belongs to Sigma_zz
		D[1][i] = c[4][4] * (lz[i] * P[0][i] + sintheta * P[2][i]) + c[3][4] * lz[i] * P[1][i];  // belongs to Tau_xz
		D[2][i] = c[3][4] * (lz[i] * P[0][i] + sintheta * P[2][i]) + c[3][3] * lz[i] * P[1][i];  // belongs to Tau_yz
	}

}

void get_char_eqn_coeff_ortho(
		double c[6][6],   /* monoclinic stiffness matrix */
		double rhocp2,    /* rho * phase_velocity^2*/
		double sin2,      /* sin^2 of the angle of incidence w.r.t. mat. coord.*/
		double a[4]) {

	double delta = 1.;

	/* three coefficients for ploynom of 2nd order in lzsq for lamb wave */
	a[0] =  (c[2][2] * c[4][4]) / delta;
	a[1] = (-c[0][2] * c[0][2] + c[0][0] * c[2][2] - 2. * c[0][2] * c[4][4] - (c[2][2] + c[4][4]) * rhocp2) / delta;
	a[2] = (c[0][0] - rhocp2) * (c[4][4] - rhocp2) / delta;
	/* lzsq for SH wave */
	a[3] = (-c[5][5] + rhocp2) / c[3][3];
}

void get_char_eqn_coeff_monocl(
		double c[6][6],   /* monoclinic stiffness matrix */
		double rhocp2,    /* rho * phase_velocity^2*/
		double sin2,      /* sin^2 of the angle of incidence w.r.t. mat. coord.*/
		double a[4]) {

//	double delta = c[0][0] * c[0][0] * c[0][0];
	double delta = 1.;
	double f1, f2, f3;
//	int i;

	a[0] = (c[2][2] * c[3][3] * c[4][4] - c[2][2] * c[3][4] * c[3][4]) / delta;

	f1 =   c[0][0] * c[2][2] * c[3][3]
	     - c[0][2] * c[0][2] * c[3][3]
	     + c[0][2] * c[2][5] * c[3][4] * 2.
	     - c[0][2] * c[3][3] * c[4][4] * 2.
	     + c[0][2] * c[3][4] * c[3][4] * 2.
	     - c[0][5] * c[2][2] * c[3][4] * 2.
	     + c[2][2] * c[4][4] * c[5][5]
	     - c[2][5] * c[2][5] * c[4][4];

	f2 = + c[2][2] * c[3][3]
	     + c[2][2] * c[4][4]
	     + c[3][3] * c[4][4]
	     - c[3][4] * c[3][4];

	a[1] = (f1 * sin2 - f2 * rhocp2) / delta;

	f1 = + c[0][0] * c[2][2] * c[5][5]
	     - c[0][0] * c[2][5] * c[2][5]
	     - c[0][0] * c[2][5] * c[3][4] * 2.
	     + c[0][0] * c[3][3] * c[4][4]
	     - c[0][0] * c[3][4] * c[3][4]
	     - c[0][2] * c[0][2] * c[5][5]      /* vorzeichen falsch, nayfeh S. 120 */
	     + c[0][2] * c[0][5] * c[2][5] * 2.
	     + c[0][2] * c[0][5] * c[3][4] * 2.
	     - c[0][2] * c[4][4] * c[5][5] * 2.
	     - c[0][5] * c[0][5] * c[2][2]
	     + c[0][5] * c[2][5] * c[4][4] * 2.;

	f2 = + c[0][0] * c[2][2]
	     + c[0][0] * c[3][3]
	     - c[0][2] * c[0][2]
	     - c[0][2] * c[4][4] * 2.
	     - c[0][5] * c[3][4] * 2.
	     + c[2][2] * c[5][5]
	     - c[2][5] * c[2][5]
	     - c[2][5] * c[3][4] * 2.
	     + c[3][3] * c[4][4]
	     - c[3][4] * c[3][4]
	     + c[4][4] * c[5][5];

	f3 = c[2][2] + c[3][3] + c[4][4];

	a[2] = (  f1 * sin2 * sin2
			- f2 * sin2 * rhocp2
			+ f3 * rhocp2 * rhocp2) / delta;

	f1 = + c[0][0] * c[4][4] * c[5][5]
	     - c[0][5] * c[0][5] * c[4][4];

	f2 = + c[0][0] * c[4][4]
	     + c[0][0] * c[5][5]
	     - c[0][5] * c[0][5]
	     + c[4][4] * c[5][5];

	f3 = c[0][0] + c[4][4] + c[5][5];

	a[3] = (  f1 * sin2 * sin2 * sin2
			- f2 * sin2 * sin2 * rhocp2
			+ f3 * sin2 * rhocp2 * rhocp2
			- rhocp2 * rhocp2 * rhocp2) / delta;

//	for(i=0; i<4; i++) {
//		fprintf(stderr,"a%d %12.3e\n",i,a[i]);
//	}
}

/*
 * compute christoffel tensor for orthotropic material and propagation direction n */
void get_ortho_christoffel_tensor(material mat, double no[3], double lambda[3][3]) {

	double c[3][3][3][3], n[3], ln;
	int i, j, k, l;

	/* get 4th order stiffness tensor. c is in GPa = 10^9 N/m^2 */
	get_ortho_stiff_tensor(mat, c);

	/* initialize christoffel tensor, normalize propagation direction */
	ln = 0;
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			lambda[i][j] = 0;
		}
		ln += no[i] * no[i];
	}
	ln = sqrt(ln);
	for (i=0; i<3; i++) {
		n[i] = no[i] / ln;
	}

	/* carry out tensor product */
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			for (k=0; k<3; k++) {
				for (l=0; l<3; l++) {
					lambda[i][l] += c[i][j][k][l] * n[j] * n[k];
				}
			}
		}
	}

	/* divide by density. rho is in g / cm^3 = 10^3 kg/m^3 = 10^3 N*s^2/m^4 */
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			lambda[i][j] /= mat.rho;
		}
	}
	
	/* lambda is now in 10^6 m^2/s^2 */
}

