/*
 * clapack.c
 *
 *  Created on: Sep 3, 2009
 *      Author: jb
 */
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <complex.h>
#include "orthodisp.h"

extern const double pi;
extern const double eps;
extern FILE *logfile;

/* Solve complex eigensystem
 *
 *   (A + lambda * B) * vl = 0
 *   vr * (A + lambda * B) = 0
 *
 *   input:
 *      A[size*size] stored row-wise: see equations above
 *      B[size*size] stored row-wise: see equations above
 *      size: dimension of the eigenproblem
 *
 *   output:
 *      lambda: eigenvalues (lambda)
 *      vl: left hand eigenvectors
 *      vr: right hand eigenvectors
 *
 *   return value:
 *      info parameter of LAPACK complex eigensystem solver zggev
 *
 *****************************************************************************/
int clapack_complex_eigensystem(complex double *A, complex double *B, int size,
		complex double *lambda, complex double *vl, complex double *vr) {

	int i, j;

	complex double AT[size*size], BT[size*size];
	complex double alpha[size], beta[size], VL[size*size], VR[size*size];
	complex double *work;
	double rwork[8*size];
	int n    = size;
	int lda  = size;
	int ldb  = size;
	int ldvl = size;
	int ldvr = size;
	int lwork;
	int info;
	char jobvl = 'V'; /* right hand ev needed 'V' / not needed 'N' */
	char jobvr = 'V'; /* left hand ev needed 'V' / not needed 'N' */

	/* copy A and B, because they are chaged in lapack function,
	 * transpose A and B because rows and columns are exchanged in lapack */
	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			AT[i*size+j] = A[j*size+i];
			BT[i*size+j] = B[j*size+i];
		}
	}

	/* call lapack function, get optimum size of workspace, solve eigensystem */
	debugmsg(DEBUG,stderr,"\n*** calling zggev ... ***\n\n");
	work = (complex double *) malloc(sizeof(complex double));
	lwork = -1;
	zggev_(&jobvl,&jobvr,&n,AT,&lda,BT,&ldb,
			alpha,beta,VL,&ldvl,VR,&ldvr,work,&lwork,rwork,&info);
	lwork = (int)(creal(work[0]));
	debugmsg(DEBUG,stderr,"    a workspace size of %d is recommended. Allocating memory...",(int)lwork);
	work = (complex double *) realloc(work,sizeof(complex double)*lwork);
	zggev_(&jobvl,&jobvr,&n,AT,&lda,BT,&ldb,
			alpha,beta,VL,&ldvl,VR,&ldvr,work,&lwork,rwork,&info);
	free(work);

	/* error checking */
	if (info != 0) {
		printf("error in zggev! info = %d\n",(int)info);
		exit((int)info);
	} else {
		debugmsg(DEBUG,stderr,"\n*** zggev terminated successfully! info = %d ***\n\n",(int)info);
	}

	/* output */
	debugmsg(DEBUG,logfile,"alpha\n");
	for (i=0; i<size; i++) {
		debugmsg(DEBUG,logfile," (%8.3lf%+8.3lfi)",creal(alpha[i]),cimag(alpha[i]));
	}
	debugmsg(DEBUG,logfile,"\n");

	debugmsg(DEBUG,logfile,"beta\n");
	for (i=0; i<size; i++) {
		debugmsg(DEBUG,logfile," (%8.3lf%+8.3lfi)",creal(beta[i]),cimag(beta[i]));
	}
	debugmsg(DEBUG,logfile,"\n\n");

	/* calculate eigenvalues, copy eigenvectors */
	for (i=0; i<size; i++) {
		lambda[i] = alpha[i] / beta[i];
		for (j=0; j<size; j++) {
			vr[i*size+j]  = VR[i*size+j];
			vl[i*size+j]  = VL[i*size+j];
		}
	}

	/* print solution: eigenvalues, eigenvectors */
	for (i=0; i<size; i++) {
		debugmsg(DEBUG,logfile,"ev%d (%8.3lf%+8.3lfi)\nvr%d",i,creal(lambda[i]),cimag(lambda[i]),i);
		for (j=0; j<size; j++) {
			debugmsg(DEBUG,logfile," (%7.3lf%+7.3lfi)",creal(vr[i*size+j]),cimag(vr[i*size+j]));
		}
		debugmsg(DEBUG,logfile,"\nvl%d",i);
		for (j=0; j<size; j++) {
			debugmsg(DEBUG,logfile," (%7.3lf%+7.3lfi)",creal(vl[i*size+j]),cimag(vl[i*size+j]));
		}
		debugmsg(DEBUG,logfile,"\n\n");
	}
	return (int) info;
}

/* Determinant of complex matrix
 *
 *   input:
 *      C[size*size] stored row-wise
 *      size: number of rows or columns of C
 *
 *   return value:
 *      determinant of C
 *
 *****************************************************************************/
complex double clapack_complex_det(complex double *C, int size) {
	int i, j, piv = 1.;
	complex double det = 1.;
	int n=size, lda = size, ipiv[size], info;
	complex double CT[size*size];

	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			CT[i*size+j] = C[j*size+i];
		}
	}
	zgetrf_(&n,&n,CT,&lda,ipiv,&info);
	for (i=0; i<size; i++) {
		if (ipiv[i] != i+1) { 		// sign from number of permutations
			det *= -CT[i*size+i];
			piv = -piv;
		} else {
			det *= CT[i*size+i];
		}
	}

	if (info == 0) {
		debugmsg(DEBUG,stdout,"\n*** determinant calculated successfully ***\n");
	} else {
		fprintf(stdout,"\n*** error while calculating determinant ***\n");
		exit(3);
	}

	return det;
}

/* Invert complex matrix
 *
 *   input:
 *      C[size*size] stored row-wise: matrix to invert
 *      size: number of rows or columns of C
 *
 *    output:
 *      CI[size*size] stored row-wise: inverse of C
 *
 *   return value:
 *      info parameter of LAPACK complex inversion routine
 *
 *****************************************************************************/
int clapack_complex_inverse(complex double *C, complex double *CI, int size) {
	int i, j;
	int n=size, lda = size, ipiv[size], info, lwork;
	complex double *work;
	complex double CT[size*size];

	/* copy original matrix */
	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			CT[i*size+j] = C[j*size+i];
		}
	}

	/* compute inverse */
	debugmsg(DEBUG,stderr,"\n*** factorize: calling zgetrf ... ***\n");
	zgetrf_(&n,&n,CT,&lda,ipiv,&info);
	debugmsg(DEBUG,stderr,"\n*** invert:  calling zgetri ... ***\n");
	work = (complex double *) malloc(sizeof(complex double));
	/* for a 6x6 matrix lwork seems to be 6. allocate 12 to be on the safe side.
	 * alternatively use following code to determine lwork synamically */
//	lwork = 2 * size;
	lwork = -1;
	zgetri_(&n,CT,&lda,ipiv,work,&lwork,&info);
	lwork = (int)(creal(work[0]));
	debugmsg(DEBUG,stderr,"    allocate recomended workspace, size %d ...\n",(int)lwork);
	work = (complex double *) realloc(work,sizeof(complex double)*lwork);
	zgetri_(&n,CT,&lda,ipiv,work,&lwork,&info);
	free(work);

	/* copy matrix back to double complex */
	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			CI[j*size+i] = CT[i*size+j];
		}
	}

	/* print message */
	if (info == 0) {
		debugmsg(DEBUG,stderr,"\n*** matrix inversion completed successfully ***\n");
	} else {
		fprintf(stderr,"\n*** error while inverting matrix ***\n");
		exit(3);
	}

	return (int)info;
}
