#define DEBUG 0

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include "orthodisp.h"

extern const double pi;
extern FILE *logfile;

/*
 * Input functions
 **********************************************************************/

/*
 *  read all parameters from input file
 **********************************************************************/

void parse_input_file(
		char *inpfn, char *logfn,
		materialDB *matdb,
		laminate *lam,
		double *sintheta, double *phi,
		double *cp_start, double *cp_end, double *cp_incr, int *cp_steps,
		double *w_start, double *w_end, double *w_incr, int *w_steps) {
	int il;

	if ( !(logfile = fopen(logfn,"w"))) {
		printf("could not open logfile for writing\nAborting...\n");
		exit(1);
	}
	matdb->size = count_materials(inpfn);
	matdb->mat = (material *) malloc(sizeof(material) * (matdb->size));
	read_materials(inpfn,matdb->mat);
	read_range(inpfn,cp_start,cp_end,cp_steps,w_start,w_end,w_steps);
	*w_start *= 1.e6; 	// need w in Hz for now
	*w_end   *= 1.e6;
	if (*w_steps > 1) {
		*w_incr  = (*w_end - *w_start) / (*w_steps - 1);
	}
	if (*cp_steps > 1) {
		*cp_incr  = (*cp_end - *cp_start) / (*cp_steps - 1);
	}
	read_waveprop(inpfn,sintheta,phi);
	lam->numlayers = count_layers(inpfn);
	lam->stacking = (layer *) malloc(sizeof(layer) * (lam->numlayers));
	read_layers(inpfn,matdb->mat,matdb->size,*phi,lam->stacking);
	lam->thickness = 0;
	for (il=0; il<(lam->numlayers); il++) {
		lam->thickness += lam->stacking[il].thickness;
	}
}

/*
 * utilities for fast-forward to a certain section and for checking
 * end of section
 */
int skip_to_section(FILE *in, char *sectname) {
	int l = strlen(sectname);
	int ret = 1;
	char line[1002];
	char *fstat = fgets(line,1002,in);

	while ((fstat != NULL) && (strncmp(line,sectname,l) != 0)) {
		fstat = fgets(line,1002,in);
	}
	if (fstat == NULL) ret = 0;
	return ret;
}

int is_section_start(char *line) {
	int i, ret = 0;
	char *section[4] = { "LAYUP", "WAVE", "RANGE", "MATERIAL" };
	for (i=0; i<4; i++) {
		if (strncmp(line,section[i],strlen(section[i])) == 0) ret = 1;
	}
	return ret;
}

/*
 * count layers of laminate and read stacking sequence
 */
int count_layers(char *filename) {
	char line[1002], name[13], *fstat;
	int nl = 0;
	double t, a;
	FILE *in = fopen(filename,"r");

	skip_to_section(in,"LAYUP");
	fstat = fgets(line,1002,in);

	while ((fstat != NULL) && (! is_section_start(line))) {
		if ((sscanf(line,"%s %lf %lf",name,&t,&a) == 3) && (name[0] != '#')) {
			debugmsg(DEBUG,stderr,"found layer of material %s\n",name);
			nl++;
		}
		fstat = fgets(line,1002,in);
	}
	fclose(in);

	debugmsg(DEBUG,stderr,"found %d layers in file %s\n",nl,filename);

	return nl;
}

int read_layers(char *filename, material *matdb, int nmat, double phi, layer *laminate) {
	char line[1002], name[13], *fstat;
	int imat, nl = 0;
	double t, a;
	FILE *in = fopen(filename,"r");

	skip_to_section(in,"LAYUP");
	fstat = fgets(line,1002,in);

	while ((fstat != NULL) && (! is_section_start(line))) {
		if ((sscanf(line,"%s %lf %lf",name,&t,&a) == 3) && (name[0] != '#')) {
			/* look for material in database */
			for (imat=0; imat<nmat; imat++) {
				if (strcmp(matdb[imat].name,name) == 0) {
					laminate[nl].mat = &(matdb[imat]);
					laminate[nl].thickness = t / 1000; /* thickness is stored in m, but specified in mm */
					laminate[nl].angle = a * pi / 180. - phi; /* phi is direction of wave propagation */
					set_layer_C(&(laminate[nl]));
					break;
				}
			}
			if (imat == nmat) {
				fprintf(stderr,"ERROR while reading LAYUP section.\nMaterial %s not found in database. Check your input file\n",name);
				exit(5);
			} else {
				debugmsg(DEBUG,stderr,"layer %2d: %12s, t=%7.3lf mm, angle=%7.1lf deg.\n",
						nl+1,name,t,a);
			}
			nl++;
		}
		fstat = fgets(line,1002,in);
	}
	fclose(in);
	return nl;
}

/*
 * read properties of incident wave
 */
int read_waveprop(char *filename, double *sintheta, double *phi) {
	char line[1002], key[3], *fstat;
	int error = 0, ipa = 0, iia = 0;
	double d, theta;
	FILE *in = fopen(filename,"r");

	skip_to_section(in,"WAVE");
	fstat = fgets(line,1002,in);

	while ((fstat != NULL) && (! is_section_start(line))) {
		if (sscanf(line,"%s %lf",key,&d) == 2) {
			if (strcmp(key,"pa") == 0) {
				/* complete stiffness matrix will be rotated by (-phi)
				 * (the negative wave propagation angle) */
				*phi = d * pi / 180.;
				ipa = 1;
			} else if (strcmp(key,"ia") == 0) {
				theta = d * pi / 180.;
				iia = 1;
			} else if (key[0] == '#') {
			} else {
				error = 1;
			}
		}
		fstat = fgets(line,1002,in);
	}
	fclose(in);

	/* check data */
	if ((error == 1) || (ipa == 0) || (iia == 0) ) {
		fprintf(stderr,"ERROR while reading WAVE section. Check your input file\n");
		exit(3);
	}
	debugmsg(DEBUG,stderr,"wave propagation angle: %10.3e deg, incident angle: %10.3e deg\n",
			*phi*180/pi,theta*180/pi);
	*sintheta = sin(theta);
	return 1;
}

/*
 * read ranges for frequency and phase velocity of dispersion diagram
 */
int read_range(char *filename,
		double *cp0, double *cp1, int *cpinc,
		double *w0,  double *w1,  int *winc) {
	char line[1002], key[3], *fstat;
	int error = 0;
	int i0;
	double d0, d1;
	FILE *in = fopen(filename,"r");

	skip_to_section(in,"RANGE");
	fstat = fgets(line,1002,in);

	*cpinc = *winc = 0;
	while ((fstat != NULL) && (! is_section_start(line))) {
		if (sscanf(line,"%s %lf %lf %d",key,&d0,&d1,&i0) == 4) {
			if (strcmp(key,"cp") == 0) {
				*cp0 = d0;
				*cp1 = d1;
				*cpinc = i0;
			} else if (strcmp(key,"fq") == 0) {
				*w0 = d0 * 2. * pi;
				*w1 = d1 * 2. * pi;
				*winc = i0;
			} else if (key[0] == '#') {
			} else {
				error = 1;
			}
		}
		fstat = fgets(line,1002,in);
	}
	fclose(in);

	/* check data */
	if ((error == 1) || (*cpinc <= 0) || (*winc <= 0)
			|| (*cp0 > *cp1) || (*cp0 < 0)
			|| ( *w0 > *w1 ) || ( *w0 < 0) ) {
		fprintf(stderr,"ERROR while reading RANGE section. Check your input file\n");
		exit(3);
	}
	debugmsg(DEBUG,stderr,"phase vel. range from %10.3e km/s to %10.3e km/s, %d steps\n",
			*cp0, *cp1, *cpinc);
	debugmsg(DEBUG,stderr,"frequency  range from %10.3e MHz  to %10.3e MHz,  %d steps\n",
			*w0/2./pi, *w1/2./pi, *winc);
	return 1;
}

/*
 * count materials and read material database
 */
int count_materials (char *filename) {
	char line[1002], name[13], *fstat;
	int nmat = 0, nisc;
	double rho, s1, s2;
	FILE *in = fopen(filename,"r");

	skip_to_section(in,"MATERIAL");
	fstat = fgets(line,1002,in);

	while ((fstat != NULL) && (! is_section_start(line))) {
		if ((sscanf(line,"%s %d %lf %lf %lf",name,&nisc,&rho,&s1,&s2) == 5) && (name[0] != '#')) {
			debugmsg(DEBUG,stderr,"found material %s with %d independent stiffness constants\n",name,nisc);
			nmat++;
		}
		fstat = fgets(line,1002,in);
	}
	fclose(in);

	debugmsg(DEBUG,stderr,"found %d materials in file %s\n",nmat,filename);

	return nmat;
}

/* too large numbers not good for precision, use 10^9 Pa for modulus
 * rho*cp*cp must have same unit as stiffness matrix */
int read_materials(char *filename, material *matdb) {
	char line[1002], name[13], *fstat;
	int imat = 0, nisc, error = 0;
	double rho, s1, s2, s3, s4, s5, s6, s7, s8, s9;
	FILE *in = fopen(filename,"r");

	skip_to_section(in,"MATERIAL");
	fstat = fgets(line,1002,in);

	while ((fstat != NULL) && (! is_section_start(line))) {
		if ((sscanf(line,"%s %d %lf %lf %lf",name,&nisc,&rho,&s1,&s2) == 5)  && (name[0] != '#')){
			switch (nisc) {
			case 2:  // isotropic material
				strncpy(matdb[imat].name,name,13);
				matdb[imat].rho  = rho;
				matdb[imat].E1   = matdb[imat].E2   = matdb[imat].E3   = s1;
				matdb[imat].nu12 = matdb[imat].nu23 = matdb[imat].nu31 = s2;
				matdb[imat].G12  = matdb[imat].G23  = matdb[imat].G31  = 0.5 * s1 / (1 + s2);
				matdb[imat].id   = imat;
				break;
			case 5:  // transversely isotropic material
				if (sscanf(line,"%s %d %lf %lf %lf %lf %lf %lf",name,&nisc,&rho,&s1,&s2,&s3,&s4,&s5) == 8) {
					strncpy(matdb[imat].name,name,13);
					matdb[imat].rho  = rho;
					matdb[imat].E1   = s1;
					matdb[imat].E2   = matdb[imat].E3   = s2;
					matdb[imat].G12  = matdb[imat].G31  = s3;
					matdb[imat].G23  = s4;
					matdb[imat].nu12 = s5;
					matdb[imat].nu23 = 0.5 * s2 / s4 - 1.0;
					matdb[imat].nu31 = s5 * s2 / s1;
					matdb[imat].id   = imat;
				} else {
					error = 1;
				}
				break;
			case 9:  // orthotropic material
				if (sscanf(line,"%s %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",name,&nisc,&rho,&s1,&s2,&s3,&s4,&s5,&s6,&s7,&s8,&s9) == 12) {
					strncpy(matdb[imat].name,name,13);
					matdb[imat].rho  = rho;
					matdb[imat].E1   = s1;
					matdb[imat].E2   = s2;
					matdb[imat].E3   = s3;
					matdb[imat].G12  = s4;
					matdb[imat].G23  = s5;
					matdb[imat].G31  = s6;
					matdb[imat].nu12 = s7;
					matdb[imat].nu23 = s8;
					matdb[imat].nu31 = s9;
					matdb[imat].id   = imat;
				} else {
					error = 1;
				}
				break;
			default:
				error = 1;
			}
			if (error != 0) {
				fprintf(stderr,"ERROR while reading material %s. Check your input file\n",name);
				exit(2);
			} else {
				debugmsg(DEBUG,stderr,"successfully read material %s from input file.\n",name);
			}
			imat++;
		}
		fstat = fgets(line,1002,in);
	}
	fclose(in);

	return imat;
}

/*
 * Output utilities
 **********************************************************************/
void debugmsg(int debuglvl, FILE *f, char *format, ...) {
	va_list ap;
	va_start(ap,format);
	if ((debuglvl > 0) && (debuglvl >= DEBUG)) {
		vfprintf(f,format,ap);
		fflush(f);
	}
}

void print_d_matrix(FILE *f, double *M, int r, int c) {
	int i, j;
	for (i=0; i<r; i++) {
		for (j=0; j<c; j++) {
			fprintf(f,"%12.3e ",M[i*c+j]);
		}
		fprintf(f,"\n");
	}
}

void print_c_matrix(FILE *f, complex double *M, int r, int c) {
	int i, j;
	for (i=0; i<r; i++) {
		for (j=0; j<c; j++) {
			fprintf(f,"%11.3e%+10.3ei  ",creal(M[i*c+j]),cimag(M[i*c+j]));
		}
		fprintf(f,"\n");
	}
}
