/*
 * material.c
 *
 *  Created on: Jun 25, 2009
 *      Author: jb
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
#include "orthodisp.h"

/*
 * Stiffness matrix, tensor transformation
 **********************************************************************/
void set_layer_C(layer *l) {
	double C0[6][6], C1[6][6], TransI[6][6], TransIT[6][6];
	double a = l->angle;

	get_ortho_stiff(*(l->mat), C0);
	get_trans_i(a,TransI);
	matmul((double *)TransI,(double *)C0,(double *)C1,6);
	transpose((double *)TransI,(double *)TransIT, 6);
	matmul((double *)C1,(double *)TransIT,(double *)(l->C),6);
}
void get_layer_prop(layer *l, double *rho, double *d, double C[6][6]) {
	int i, j;
	material *m = l->mat;
	*d = l->thickness;
	*rho = m->rho;

	for (i=0; i<6; i++) {
		for (j=0; j<6; j++) {
			C[i][j] = l->C[i][j];
		}
	}
}

/*
 * stiffness matrix (tensor notation), cut & pasted from inverse of
 * compliance matrix solved by Mathematica
 */
void get_ortho_stiff_tensor(material mat, double c[3][3][3][3]) {
	
	double C[6][6];
	int i, j, k, l, M, N;
	
	get_ortho_stiff(mat, C);
	
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			M = (i == j ? i : 6-i-j);
			for (k=0; k<3; k++) {
				for (l=0; l<3; l++) {
					N = (k == l ? k : 6-k-l);
					c[i][j][k][l] = C[M][N];
				}
			}
		}
	}
}

/*
 * stiffness matrix (voigt notation), cut & pasted from inverse of
 * compliance matrix solved by Mathematica
 */
void get_ortho_stiff(material mat, double C[6][6]) {

	int i, j;

	double E1 = mat.E1;
	double E2 = mat.E2;
	double E3 = mat.E3;
	double G12 = mat.G12;
	double G23 = mat.G23;
	double G31 = mat.G31;
	double nu12 = mat.nu12;
	double nu23 = mat.nu23;
	double nu31 = mat.nu31;

	double Delta = 1. - 2.*nu12*nu23*nu31
		- (E2*nu12*nu12)/E1 - (E3*nu23*nu23)/E2 - (E1*nu31*nu31)/E3;

	for (i=0; i<6; i++) {
		for (j=0; j<6; j++) {
			C[i][j] = 0.;
		}
	}

	C[0][0] = (E1 - (E1*E3*nu23*nu23)/E2) / Delta;
	C[1][1] = (E2 - (E1*E2*nu31*nu31)/E3) / Delta;
	C[2][2] = (E3 - (E2*E3*nu12*nu12)/E1) / Delta;
	C[0][1] = C[1][0] = (E2*nu12 + E1*nu23*nu31) / Delta;
	C[1][2] = C[2][1] = (E3*nu23 + E2*nu12*nu31) / Delta;
	C[0][2] = C[2][0] = (E1*nu31 + E3*nu12*nu23) / Delta;
	C[3][3] = G23;
	C[4][4] = G31;
	C[5][5] = G12;
}


void get_trans(double phi, double T[6][6]) {
	int i, j;
	double c = cos(phi);
	double s = sin(phi);
	double cc = c*c, ss = s*s, cs = c*s;

	for (i=0; i<6; i++) {
		for (j=0; j<6; j++) {
			T[i][j] = 0.;
		}
	}
	T[0][0] = cc;
	T[0][1] = ss;
	T[0][5] = 2.*cs;
	T[1][0] = ss;
	T[1][1] = cc;
	T[1][5] = -2.*cs;
	T[2][2] = 1.;
	T[3][3] = c;
	T[3][4] = -s;
	T[4][3] = s;
	T[4][4] = c;
	T[5][0] = -cs;
	T[5][1] = cs;
	T[5][5] = cc - ss;
}

void get_trans_i(double phi, double TI[6][6]) {
	int i, j;
	double c = cos(phi);
	double s = sin(phi);
	double cc = c*c, ss = s*s, cs = c*s;

	for (i=0; i<6; i++) {
		for (j=0; j<6; j++) {
			TI[i][j] = 0.;
		}
	}
	TI[0][0] = cc;
	TI[0][1] = ss;
	TI[0][5] = -2.*cs;
	TI[1][0] = ss;
	TI[1][1] = cc;
	TI[1][5] = 2.*cs;
	TI[2][2] = 1.;
	TI[3][3] = c;
	TI[3][4] = s;
	TI[4][3] = -s;
	TI[4][4] = c;
	TI[5][0] = cs;
	TI[5][1] = -cs;
	TI[5][5] = cc - ss;
}

int is_monocl(double C[6][6]) {
	int i, ret = 1;
	double dmincii = 1.e200, c05, c15, c25, c34;
	double epsc = 1.e-07;

	/* compare monoclinic components to min(C00,C11,C22) */
	for (i=0; i<3; i++) {
		dmincii = (C[i][i] < dmincii ? C[i][i] : dmincii);
	}
	c05 = fabs(C[0][5]/dmincii);
	c15 = fabs(C[1][5]/dmincii);
	c25 = fabs(C[2][5]/dmincii);
	c34 = fabs(C[3][4]/dmincii);

	if (
			(c05 < epsc) &&
			(c15 < epsc) &&
			(c25 < epsc) &&
			(c34 < epsc)
			) {
		ret = 0;
	}
	return ret;
}
