/*
 * math.c
 *
 *  Created on: Jun 25, 2009
 *      Author: jb
 */
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <complex.h>
#include "orthodisp.h"

extern const double pi;
extern const double eps;
extern FILE *logfile;

/*
 * reduce singular system of equation defined by matrix (M * x = 0)
 * to a system of equation of size-1 by deleting row r and storing
 * negative column c of M in the right hand side: LHS * y = Rhs
 **********************************************************************/
void reduce_singular_soe(complex double *M, int size, int rD, int cR,
		complex double *LHS, complex double *Rhs) {
	int rM, cM, rL, cL;

	for (rM=rL=0; rM<size; rM++) {
		if (rM != rD) {
			for (cM=cL=0; cM<size; cM++) {
				if (cM != cR) {
					LHS[rL*(size-1)+cL] = M[rM*size+cM];
					cL++;
				} else {
					Rhs[rL] = -M[rM*size+cM];
				}
			}
			rL++;
		}
	}
}

/*
 * find zeros: analyse vector c for change in sign in either real
 * or complex part.
 **********************************************************************/
void find_zeros_c(complex double *c, int size, int *s) {
	int i;
	double eps = 1.e-99;
	double *dr = malloc(sizeof(double)*size);
	double *di = malloc(sizeof(double)*size);
	double *re = malloc(sizeof(double)*size);
	double *im = malloc(sizeof(double)*size);
	int *sr = malloc(sizeof(int)*size);
	int *si = malloc(sizeof(int)*size);

	for (i=0; i<size; i++) {
		re[i] = (fabs(creal(c[i])) >= fabs(cimag(c[i])) ? creal(c[i]) : 0.);
		im[i] = (fabs(cimag(c[i])) >= fabs(creal(c[i])) ? cimag(c[i]) : 0.);
		dr[i] = 0.;
		di[i] = 0.;
		sr[i] = 0;
		si[i] = 0;
	}
	// 1st filter: evaluate sign just from values in orig vector,
	// compute derivatives
	for (i=2; i<size-2; i++) {
		if ((fabs(re[i-1]) > eps) && (fabs(re[i]) > eps)) {
			dr[i] = re[i] - re[i-1];
		} else {
			dr[i] = 0.;
		}
		if (       (fabs(re[i-2]) > eps)
				&& (fabs(re[i-1]) > eps)
				&& (fabs(re[i])   > eps)
				&& (fabs(re[i+1]) > eps)
				&& (sign(re[i-2]) == sign(re[i-1]))
				&& (sign(re[i-1]) != sign(re[i])  )
				&& (sign(re[i])   == sign(re[i+1])) ) {
			sr[i] = 1;
		}
		if ((fabs(im[i-1]) > eps) && (fabs(im[i]) > eps)) {
			di[i] = im[i] - im[i-1];
		} else {
			di[i] = 0.;
		}
		if (       (fabs(im[i-2]) > eps)
				&& (fabs(im[i-1]) > eps)
				&& (fabs(im[i])   > eps)
				&& (fabs(im[i+1]) > eps)
				&& (sign(im[i-2]) == sign(im[i-1]))
				&& (sign(im[i-1]) != sign(im[i])  )
				&& (sign(im[i])   == sign(im[i+1])) ) {
			si[i] = 1;
		}
	}

	// 2nd filter: eliminate poles by looking at derivative before and
	// after change of sign. derivative and value must have opposite sign
	s[0] = s[size-1] = 0;
	for (i=1; i<size-1; i++) {
		if ( (sr[i] != 0) && ((sign(re[i+1]) != sign(dr[i+1])) || (sign(re[i-1]) == sign(dr[i-1]))) ) {
			sr[i] = 0;
		}
		if ( (si[i] != 0) && ((sign(im[i+1]) != sign(di[i+1])) || (sign(im[i-1]) == sign(di[i-1]))) ) {
			si[i] = 0;
		}
		// this is the result vector;
		s[i] = sr[i] + si[i];
	}

	for (i=0; i<size; i++) {
		debugmsg(0,stderr,"r %11.3e %11.3e dr %11.3e sr %1d    i %11.3e %11.3e di %11.3e si %1d    s %1d\n",
				creal(c[i]),re[i],dr[i],sr[i],cimag(c[i]),im[i],di[i],si[i],s[i]);
	}

	free(sr);
	free(si);
	free(dr);
	free(di);
	free(re);
	free(im);
	return;
}

/*
 * evaluate sign of double
 **********************************************************************/
int sign(double n) {
	if (n < 0) return -1;
	if (n > 0) return 1;
	return 0;
}

/* multiply 2 square double matrices
 * NOTE: matmul(a,b,a) for a := a*b is NOT allowed
 **********************************************************************/
void matmul(double *a, double *b, double *c, int size) {
	int i, j, k;

	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			c[i*size+j] = 0.;
		}
	}
	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			for (k=0; k<size; k++) {
				c[i*size+j] += a[i*size+k] * b[k*size+j];
			}
		}
	}
}

/* multiply 2 square complex matrices
 * copy matrix a to a0, to allow complex_matmul(a,b,a) for a := a*b
 **********************************************************************/
void complex_matmul(
		complex double *a,
		complex double *b,
		complex double *c, int size) {
	int i, j, k;
	complex double a0[size*size];

	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			a0[i*size+j] = a[i*size+j];
			c[i*size+j] = 0.;
		}
	}
	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			for (k=0; k<size; k++) {
				c[i*size+j] += a0[i*size+k] * b[k*size+j];
			}
		}
	}
}

/* multiply square complex matrix with vector
 **********************************************************************/
void complex_vecmul(
		complex double *a,
		complex double *b,
		complex double *c, int size) {
	int i, j;

	for (i=0; i<size; i++) {
		c[i] = 0.;
		for (j=0; j<size; j++) {
			c[i] += a[i*size+j] * b[j];
		}
	}
}

/* transpose 2 square double matrices
 **********************************************************************/
void transpose(double *a, double *b, int size) {
	int i, j;

	for (i=0; i<size; i++) {
		for (j=0; j<size; j++) {
			b[i*size+j] = a[j*size+i];
		}
	}
}

/* find roots of 2nd order polynom
 **********************************************************************/
void roots_of_poly024_ortho(
		double a[4],
		complex double lz[6])
{
	double a0x2 = 2. * a[0], a1 = a[1], a2x2 = 2. * a[2], a3 = a[3];

	complex double f = csqrt(a1 * a1 - a0x2 * a2x2);
	complex double lzsq0, lzsq1, lzsq2;

	lzsq0 = (-a1 - f) / a0x2;  // lamb 1
	lzsq1 = (-a1 + f) / a0x2;  // lamb 2
	lzsq2 = a3;  // hs-wave

	lz[0] =  csqrt(lzsq0);
	lz[1] =  csqrt(lzsq1);
	lz[2] =  csqrt(lzsq2);
	lz[3] = -csqrt(lzsq0);
	lz[4] = -csqrt(lzsq1);
	lz[5] = -csqrt(lzsq2);

	return;
}

/* find roots of 3rd order polynom
 **********************************************************************/
void roots_of_poly0246(
		double a[4],
		complex double lz[6])
{
	const double twop23 = pow(2.,2./3);
	const double twop13 = pow(2.,1./3);
	const complex double omisqrt3 = 1. - I * sqrt(3.);
	const complex double isqrt3p1 = 1. + I * sqrt(3.);

	double a0 = a[0], a1 = a[1], a2 = a[2], a3 = a[3];
	double a0sq = a0*a0;
	double a1sq = a1*a1;
	double a1tpo3 = a1sq*a1;

	double g = 2. * a1tpo3 - 9. * a0*a1*a2 + 27. * a0sq*a3;
	complex double f = cpow(-g + csqrt(-4.*cpow(a1sq - 3.*a0*a2,3.) + g*g),1./3.);
	complex double of;
	complex double lzsq0, lzsq1, lzsq2;
	complex double tmp;

	of = (fabs(g) < eps ? 0. : 1./f);

	lzsq0 = (-2.*a1 + 2.*(a1sq - 3.*a0*a2)*of*twop13 + f*twop23)/(6.*a0);
	lzsq1 = (-4.*a1 - 2.*(a1sq - 3.*a0*a2)*isqrt3p1*of*twop13 - f*omisqrt3*twop23)/(12.*a0);
	lzsq2 = -(4.*a1 + 2.*(a1sq - 3.*a0*a2)*of*omisqrt3*twop13 + f*isqrt3p1*twop23)/(12.*a0);
	if (creal(lzsq1) > creal(lzsq0)) {
		tmp = lzsq1;
		lzsq1 = lzsq0;
		lzsq0 = tmp;
	}
	if (creal(lzsq2) > creal(lzsq1)) {
		tmp = lzsq2;
		lzsq2 = lzsq1;
		lzsq1 = tmp;
	}
	if (creal(lzsq1) > creal(lzsq0)) {
		tmp = lzsq1;
		lzsq1 = lzsq0;
		lzsq0 = tmp;
	}

	/* lzsq should now be ordered according to their size */
	debugmsg(DEBUG,logfile,"lzsq1: %12.3e%+12.3e i", creal(lzsq0), cimag(lzsq0));
	debugmsg(DEBUG,logfile,"lzsq2: %12.3e%+12.3e i", creal(lzsq1), cimag(lzsq1));
	debugmsg(DEBUG,logfile,"lzsq3: %12.3e%+12.3e i\n", creal(lzsq2), cimag(lzsq2));

	lz[0] =  csqrt(lzsq0);
	lz[3] = -csqrt(lzsq0);
	if (cimag(lz[0]) < 0) {
		lz[0] = -lz[0];
		lz[3] = -lz[3];
	}
	lz[1] =  csqrt(lzsq1);
	lz[4] = -csqrt(lzsq1);
	if (cimag(lz[1]) < 0) {
		lz[1] = -lz[1];
		lz[4] = -lz[4];
	}

	lz[2] =  csqrt(lzsq2);
	lz[5] = -csqrt(lzsq2);
	if (cimag(lz[2]) < 0) {
		lz[2] = -lz[2];
		lz[5] = -lz[5];
	}

	return;
}
