/*
 * Program displam: Compute dispersion diagrams for Lamb waves in
 *                  laminated plates composed of transversely isotropic
 *                  or orthotropic layers.
 *
 * Copyright (c) 2009 German Aerospace Center, DLR, Institute FA.
 *
 * Theory
 *       - Nayfeh 1991: Wave propagation in layered anisotropic media
 *
 *       - Rokhlin, Wang 2002: Stable recursive algorithm for elastic
 *         wave propagation in layered anisotropic media: stiffness
 *         matrix method
 *
 *       - Wang, Rokhlin 2002: Stable reformulation of transfer matrix
 *         method for wave propagation in layered anisotropic media
 *
 * Created: Jens Baaran (jb), August - September 2009
 */

#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <complex.h>

#include "orthodisp.h"

/* --------------------------------------------- forward declarations */
const double pi = M_PI;
const double eps = 1.e-20; // used for roots of 3rd order polynom (math.c)
FILE *logfile;          // logfile for debugging

/* ------------------------------------------------------------- main */
int main(int argc, char *argv[]) {

	/* material properties */
	materialDB matdb;
	int *imat;
	char bulkfn[21];

	/* laminate properties, stiffness matrix */
	int il;
	laminate lam;
	complex double KT[6][6];

	/* wave properties */
	double w, cp, waveno;
	int icp, iw;
	double sintheta; 	/* angle of incident wave w.r.t z-axis */
	double phi;         /* in-plane wave prop. angle */

	/* iteration parameters for phase velocity and circ. frequency */
	double cp_start, cp_end, cp_incr, w_start, w_end, w_incr;
	int    cp_steps, w_steps;
	double *cpvec;

	/* parameters related to residuum of dispersion equation */
	complex double *resm;
	int *sgn;

	/* displacement vectors on top and bottom of the laminate */
	complex double displ_u[3], displ_l[3];
	int i, idisp, iflag;
	double maxd;

	/* output files for dispersion data, input file name */
	char inpfn[1002]="sample.inp";
	FILE *os, *in;

	/* if there is a command line argument it's an input filename */
	if (argc > 1) {
		strncpy(inpfn,argv[1],1002);
	}
	if (! (in = fopen(inpfn,"r"))) {
		debugmsg(1,stderr,"Error: can't open input file %s for reading.\n",inpfn);
		debugmsg(1,stderr,"Usage: %s [ input_file ]\n",argv[0]);
		debugmsg(1,stderr,"Default \"input_file\" name is sample.inp, if no\n");
		debugmsg(1,stderr,"input file is specified on the command line.\n");
		exit(2);
	} else {
		fclose(in);
	}

	/* open log file */
	if ( !(logfile = fopen("log.txt","w"))) {
		printf("could not open logfile for writing\nAborting...\n");
		exit(1);
	}

	/* parse input file */
	parse_input_file(inpfn,"log.txt",
			&matdb,	&lam, &sintheta, &phi,
			&cp_start, &cp_end, &cp_incr, &cp_steps,
			&w_start, &w_end, &w_incr, &w_steps);

	/* material symmetry is not allowed, check all layers of the laminate */
	for (il=0; il<lam.numlayers; il++) {
		if (! is_monocl(lam.stacking[il].C)) {
			debugmsg(1,stderr,"ERROR: material symmetry encountered for layer %d.\n",il+1);
			debugmsg(1,stderr,"Material symmetry is not allowed. Please modify your input file.\n",il);
			exit(5);
		}
	}

	/* allocate */
	resm  = (complex double *) malloc(sizeof (complex double) * cp_steps);
	sgn   = (int *) malloc(sizeof (int) * cp_steps);
	cpvec = (double *) malloc(sizeof (double) * cp_steps);
	imat  = (int *) malloc(sizeof (int) * matdb.size);

	/* analyze bulk wave velocities */
	for (i=0; i<matdb.size; i++) {
		imat[i] = 0;
	}
	for (i=0; i<lam.numlayers; i++) {
		imat[lam.stacking[i].mat->id]++;
	}
	for (i=0; i<matdb.size; i++) {
		if (imat[i] > 0) {
			snprintf(bulkfn,20,"%s.bulk",matdb.mat[i].name);
			print_ortho_bulk_wave(bulkfn,matdb.mat[i]);
		}
	}

	/* open output file for dispersion diagram and print header */
	if ( !(os = fopen("disp.txt","w"))) {
		printf("could not open file for writing\nAborting...\n");
		exit(1);
	}
	fprintf(os,"fd/(MHz*mm)\tcp/(km/sec)\tw/(1/sec) \tk/(1/m)    \tmode");
	for (i=0; i<3; i++) {
		fprintf(os,"\t\t  uu%dr  \t  uu%di  ",i,i);
	}
	for (i=0; i<3; i++) {
		fprintf(os,"\t\t    ul%dr   \t    ul%di   ",i,i);
	}
	fprintf(os,"\n");

	/* loops over cp and w */
	w = w_start;
	for (iw = 0; iw < w_steps; iw++) {
		cp = cp_start;

		for (icp = 0; icp < cp_steps; icp++) {

			/* wave number must have dimension 1/m */
			waveno = w / cp / 1000;

			/* calculate total stiffness matrix of the laminate */
			wang_total_stiffness_matrix (lam.stacking, lam.numlayers, cp, waveno, sintheta, KT);

			/* residuum is determinant of 6x6 stiffness matrix */
			resm[icp] = clapack_complex_det((complex double *)KT,6);

			debugmsg(DEBUG,logfile,"cp %6.3lf   w %6.3lf   sres = %11.3e%+10.3ei\n",
					cp, w/1.e6, creal(resm[icp]), cimag(resm[icp]));

			/* increment phase velocity */
			cp += cp_incr;

		}

		/* analyse residuum, find zeros, omit poles */
		find_zeros_c(resm,cp_steps,sgn);

		/* iterate to accurate cp */
		find_cps(&lam, w, sintheta, sgn, cp_steps, cp_start, cp_incr, cpvec);

		/* print wave parameters of zeros to file. these are points in dispersion diagram */
		for (icp=0; icp < cp_steps; icp++) {
			if (cpvec[icp] != 0) {
				cp = cpvec[icp];
				waveno = w / cp / 1000;

				/* print wave parameters to output file (dispersion diagram) */
				fprintf(os,"%15.8e\t%15.8e\t%15.8e\t%15.8e", w/2/pi*(lam.thickness)/1000, cp, w/1000000, waveno);

				/* calculate total stiffness matrix again */
				wang_total_stiffness_matrix(lam.stacking, lam.numlayers, cp, waveno, sintheta, KT);

				/* calculate polarisation vector for singular KT
				 * displacement are normalized w.r.t ??? */
				wang_surface_displacement(KT, displ_u, displ_l);

				/* symmetric, antimetric, or hs mode? */
				/* find max displacement component in disp_u */
				maxd = 0;
				iflag = -1;
				idisp = -1;
				for (i=0; i<3; i++) {
					if (fabs(creal(displ_u[i])) > maxd) {
						maxd = fabs(creal(displ_u[i]));
						idisp = i;
						iflag = 0;
					}
					if (fabs(cimag(displ_u[i])) > maxd) {
						maxd = fabs(cimag(displ_u[i]));
						idisp = i;
						iflag = 1;
					}
				}

				/* if y-displacement is largest component it's a quasi-HS wave */
				if (idisp == 1) {
					fprintf(os,"\tH");
				} else {
					/* otherwise compare z-displacement on top & bottom */
					if (sign(creal(displ_u[2])) == sign(creal(displ_l[2]))) {
						fprintf(os,"\tS");
					} else {
						fprintf(os,"\tA");
					}
				}

				/* print polarisation vectors at laminate surface to output file */
				for (i=0; i<3; i++) {
					fprintf(os,"\t\t%11.4e\t%+11.4e",creal(displ_u[i]),cimag(displ_u[i]));
				}
				for (i=0; i<3; i++) {
					fprintf(os,"\t\t%11.4e\t%+11.4e",creal(displ_l[i]),cimag(displ_l[i]));
				}
				fprintf(os,"\n");
			}
		}

		/* increment circular frequency */
		w += w_incr;
	}

	/* close files, free memory */
	fclose(os);
	fclose(logfile);
	free(matdb.mat);
	free(lam.stacking);
	free(resm);
	free(sgn);
	free(imat);
	free(cpvec);

	return 0;
}
