/*
 * wang.c
 *
 *  Created on: Sep 2, 2009
 *      Author: jb
 */
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <complex.h>
#include "orthodisp.h"

extern FILE *logfile;
extern const double eps;

/*  calculate stiffness matrix for a laminate
 *
 *  theory: wang, rokhlin "Stable reformulation of transfer matrix method ...", 2002
 *
 *****************************************************************************/
void wang_total_stiffness_matrix (layer *laminate, int nl, double cp,
		double waveno, double sintheta, complex double KT[6][6]) {

	/* coefficients of characteristic polynom in lzsq, and six roots
	 * of the characteristic equation, corresponding to the vertical
	 * wave number components of the six partial waves */
	double a[4];
	complex double lz[6];

	/* partial stiffness matrix */
	complex double KA[6][6];

	/* material parameters */
	double rho, d, rhocp2;

	/* iteration counters */
	int il, i, j;

	/* material properties */
	rho = laminate[0].mat->rho;
	d = laminate[0].thickness;
	rhocp2 = rho * cp * cp;

	/* calc lzs, stiffness matrix method from rokhlin wang 2002 */
	get_char_eqn_coeff_monocl(laminate[0].C, rhocp2, sintheta*sintheta, a);
	roots_of_poly0246(a,lz);
	wang_layer_stiffness_matrix(laminate[0].C, rhocp2, waveno, d, sintheta, lz, KT);

	/* loop over layers, build total stiffness matrix */
	for (il=1; il<nl; il++) {
		/* old total stiffness matrix is new stiffness matrix of upper
		 * laminate */
		for (i=0; i<6; i++) {
			for (j=0; j<6; j++) {
				KA[i][j] = KT[i][j];
			}
		}

		/* material properties */
		rho = laminate[il].mat->rho;
		d = laminate[il].thickness;
		rhocp2 = rho * cp * cp;

		/* calc lzs, stiffness matrix method from rokhlin wang 2002 */
		get_char_eqn_coeff_monocl(laminate[il].C, rhocp2, sintheta*sintheta, a);
		roots_of_poly0246(a,lz);
		wang_add_layer_to_stiffness_matrix(laminate[il].C,rhocp2,waveno,d,sintheta,lz,KA,KT);
	}
}

/*
 *  wang, rokhlin "Stable reformulation of transfer matrix method ...", 2002
 *
 *  stiffness matrix method.
 *  build total stiffness matrix out of
 *   - stiffness matrix of lower laminate
 *   - and properties of actual layer
 *
 *****************************************************************************/
void wang_add_layer_to_stiffness_matrix (
		double c[6][6],
		double rhocp2,
		double k,
		double d,
		double sintheta,
		complex double lz[6],
		complex double KA[6][6],  // stiffness matrix of lower laminate (A)
		complex double KT[6][6])  // stiffness matrix of lower laminate (A) + actual ply (B)
{
	int i, j;
	complex double KB[6][6];      // stiffness matrix of actual ply
	complex double KA11[3][3], KA12[3][3], KA21[3][3], KA22[3][3];
	complex double KB11[3][3], KB12[3][3], KB21[3][3], KB22[3][3];
	complex double KT11[3][3], KT12[3][3], KT21[3][3], KT22[3][3];
	complex double KB11mKA22[3][3], KB11mKA22I[3][3];
	complex double Tmp1[3][3], Tmp2[3][3];

	/* get stiffness matrix of actual ply B */
	wang_layer_stiffness_matrix(c,rhocp2,k,d,sintheta,lz,KB);

	/* populate submatrices */
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			KA11[i][j]      = KA[i][j];
			KA12[i][j]      = KA[i][j+3];
			KA21[i][j]      = KA[i+3][j];
			KA22[i][j]      = KA[i+3][j+3];

			KB11[i][j]      = KB[i][j];
			KB12[i][j]      = KB[i][j+3];
			KB21[i][j]      = KB[i+3][j];
			KB22[i][j]      = KB[i+3][j+3];

			KB11mKA22[i][j] = KB11[i][j] - KA22[i][j];
		}
	}

	/* invert submatrix */
	clapack_complex_inverse((complex double *)KB11mKA22,(complex double *)KB11mKA22I,3);

	/* build total stiffness matrix: see equation 21 in Rokhlin, Wang, 2002 */
	complex_matmul((complex double *)KA12,(complex double *)KB11mKA22I,(complex double *)Tmp1,3);
	complex_matmul((complex double *)KB21,(complex double *)KB11mKA22I,(complex double *)Tmp2,3);

	complex_matmul((complex double *)Tmp1,(complex double *)KA21,(complex double *)KT11,3);
	complex_matmul((complex double *)Tmp1,(complex double *)KB12,(complex double *)KT12,3);
	complex_matmul((complex double *)Tmp2,(complex double *)KA21,(complex double *)KT21,3);
	complex_matmul((complex double *)Tmp2,(complex double *)KB12,(complex double *)KT22,3);

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			KT[i][j]     = KA11[i][j] + KT11[i][j];
			KT[i][j+3]   =            - KT12[i][j];
			KT[i+3][j]   =              KT21[i][j];
			KT[i+3][j+3] = KB22[i][j] - KT22[i][j];
		}
	}
}

/*  calculate stiffness matrix for a single layer
 *
 *  wang, rokhlin "Stable reformulation of transfer matrix method ...", 2002
 *
 *****************************************************************************/
void wang_layer_stiffness_matrix (
		double c[6][6],
		double rhocp2,
		double k,
		double d,
		double sintheta,
		complex double lz[6],
		complex double K[6][6])
{
	int i, j;
	complex double P[3][6], D[3][6];
	complex double Pm[3][3], Pp[3][3], Dm[3][3], Dp[3][3];
	complex double H[3][3], pH[3][3],mH[3][3];
	complex double DDH[6][6], PPH[6][6], PPHI[6][6];
	complex double tmp;

	/* Note: lzs must be sorted properly already!
	 * obtain polarisation vectors for displacements (V,W) and stress (D) */
	get_polarisation(c, rhocp2, sintheta, lz, P, D);

	/* debug output */
	if (DEBUG) {
		debugmsg(DEBUG,logfile,"Wang lz\n");
		for (i=0; i<6; i++) debugmsg(DEBUG,logfile,"   %11.3e%+10.3ei",creal(lz[i]),cimag(lz[i]));
		debugmsg(DEBUG,logfile,"Wang P\n");
		print_c_matrix(logfile,(complex double *)P,3,6);
		debugmsg(DEBUG,logfile,"Wang D\n");
		print_c_matrix(logfile,(complex double *)D,3,6);
	}
	debugmsg(DEBUG,logfile,"kd = %11.3e\n",k*d);
	for (i = 0; i < 6; i++) {
		tmp = cexp(I * d * k * lz[i]);
		debugmsg(DEBUG,logfile,"lz%d: %11.3e%+10.3ei, e^(ikdlz): %11.3e%+10.3ei\n",i,creal(lz[i]),cimag(lz[i]),creal(tmp),cimag(tmp));
	}

	/* populate matrices */
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			Pp[i][j] = P[i][j];
			Pm[i][j] = P[i][j+3];
			Dp[i][j] = D[i][j];
			Dm[i][j] = D[i][j+3];
			H[i][j] = 0.;
		}
		H[i][i] = cexp(I * d * k * lz[i]);
	}

	if (DEBUG) {
		debugmsg(DEBUG,logfile,"Wang Pp\n");
		print_c_matrix(logfile,(complex double *)Pp,3,3);
		debugmsg(DEBUG,logfile,"Wang Pm\n");
		print_c_matrix(logfile,(complex double *)Pm,3,3);
		debugmsg(DEBUG,logfile,"Wang Dp\n");
		print_c_matrix(logfile,(complex double *)Dp,3,3);
		debugmsg(DEBUG,logfile,"Wang Dm\n");
		print_c_matrix(logfile,(complex double *)Dm,3,3);
		debugmsg(DEBUG,logfile,"Wang H\n");
		print_c_matrix(logfile,(complex double *)H,3,3);
	}

	complex_matmul((complex double *)Dp,(complex double *)H,(complex double *)pH,3);
	complex_matmul((complex double *)Dm,(complex double *)H,(complex double *)mH,3);
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			DDH[i][j]     = Dm[i][j];
			DDH[i][j+3]   = pH[i][j];
			DDH[i+3][j]   = mH[i][j];
			DDH[i+3][j+3] = Dp[i][j];
		}
	}

	complex_matmul((complex double *)Pp,(complex double *)H,(complex double *)pH,3);
	complex_matmul((complex double *)Pm,(complex double *)H,(complex double *)mH,3);
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			PPH[i][j]     = Pm[i][j];
			PPH[i][j+3]   = pH[i][j];
			PPH[i+3][j]   = mH[i][j];
			PPH[i+3][j+3] = Pp[i][j];
		}
	}

	/* build stiffness matrix K for the layer */
	clapack_complex_inverse((complex double *)PPH,(complex double *)PPHI,6);
	complex_matmul((complex double *)DDH,(complex double *)PPHI,(complex double *)K,6);

	debugmsg(DEBUG,logfile,"Wang K\n");
	if (DEBUG) {
		print_c_matrix(logfile,(complex double *)K,6,6);
	}

	return;
}

/*
 * find phase velocities cp for propagating wave with fixed frequency.
 * determine cp with greater accuracy than cp_step by iteration.
 * in each step evaluate determinant of total stiffness matrix, which
 * is zero for the correct value of cp.
 **********************************************************************/
void find_cps(
		laminate *lam,
		double w, double sintheta,
		int *sgn, int size,
		double cp_start, double cp_incr,
		double *cp) {
	int icp, iflag, iter;
	double waveno, cp0, cp1, cpn, d0r, d1r, dnr;
	complex double d0, d1, dn, KT[6][6];

	const double eps_cp = 1.e-2;
	const int maxiter = 100;

	static int user_warned = 0;

	cp[0] = 0.;

	/* zeros of det(KT) with KT = f(cp) have a non-zero sgn */
	for (icp=1; icp < size-1; icp++) {
		cp[icp] = 0.;
		if (sgn[icp] != 0) {

			/* det(KT) on one side of the cp axis */
			cp0 = cp_start + (icp-1) * cp_incr;
			waveno = w / cp0 / 1000;
			wang_total_stiffness_matrix(lam->stacking, lam->numlayers, cp0, waveno, sintheta, KT);
			d0 = clapack_complex_det((complex double *)KT,6);

			/* det(KT) on the other side of the cp axis */
			cp1 = cp_start + (icp+1) * cp_incr;
			waveno = w / cp1 / 1000;
			wang_total_stiffness_matrix(lam->stacking, lam->numlayers, cp1, waveno, sintheta, KT);
			d1 = clapack_complex_det((complex double *)KT,6);

			/* either real part or imaginary part of det(KT) must be 0. set flag and initialize */
			if (creal(d0) != creal(d1)) {
				iflag = 0;
				d0r = creal(d0);
				d1r = creal(d1);
			} else if (cimag(d0) != cimag(d1)) {
				iflag = 1;
				d0r = cimag(d0);
				d1r = cimag(d1);
			} else {
				fprintf(stderr,"FATAL ERROR: characteristic function not zero in find_cps. Aborting...\n");
				exit(6);
			}
			debugmsg(DEBUG,logfile,"iflag = %d    cp0 = %10.3e   d0 = %11.3e%+10.3ei   cp1 = %10.3e   d1 = %11.3e%+10.3ei\n",
					iflag, cp0, d0, cp1, d1);

			/* loop to find real or imaginary part of det(KT) < eps */
			iter = 0;
			do {
				/* secant method */
				//cpn = cp0 + (cp1 - cp0) / (d0 - d1) * d0;

				/* bisection method */
				cpn = (cp1 + cp0) / 2.;

				waveno = w / cpn / 1000;
				wang_total_stiffness_matrix(lam->stacking, lam->numlayers, cpn, waveno, sintheta, KT);
				dn = clapack_complex_det((complex double *)KT,6);
				dnr = (iflag == 0 ? creal(dn) : cimag(dn));
				if (sign(dnr) == sign(d0)) {
					cp0 = cpn;
					d0r = dn;   // not necessary for bisection method
				} else {
					cp1 = cpn;
					d1r = dn;   // not necessary for bisection method
				}
				iter++;
				debugmsg(DEBUG,logfile,"    i = %2d   cp0 = %17.10e   d0r = %11.3e    cp1 = %17.10e   d1r = %11.3e   |dnr| = %10.3e\n",
						iter, cp0, d0r, cp1, d1r, fabs(dnr));
			} while ((iter < maxiter) && (fabs(dnr) > eps_cp));

			/* if no convergence: warn user */
			if (fabs(dnr) > eps_cp) {
				if (! user_warned) {
					fprintf(stderr,"WARNING: no convergence for displacements uu and ul. Details in log.txt.\n");
					user_warned = 1;
				}
				debugmsg(1,logfile,"|KT| is %10.3e, threshold is %10.3e, ", fabs(dnr), eps_cp);
				debugmsg(1,logfile,"uu and ul may be inaccurate at w = %17.10e, %17.10e < cp < %17.10e\n",
						w, cp0, cp1);
			} else {
				debugmsg(DEBUG,logfile,"w = %17.10e   cp = %17.10e   |KT| = %11.3e%+10.3ei\n", w, cpn, creal(dn),cimag(dn));
			}

			/* found cp that satisfies det(KT) < eps */
			cp[icp] = cpn;
		}
	}
	cp[size-1] = 0.;
}


/* calculate displacement vector on upper (displ_u) and lower (displ_l)
 * laminate surface from
 *
 *     KT * | displ_u | = | sigma_u | = | 0 |
 *          | displ_l |   | sigma_l |   | 0 |
 *
 * The determinant of KT is 0, that means KT is singular */
void wang_surface_displacement(
		complex double KT[6][6],
		complex double displ_u[3],
		complex double displ_l[3]) {
	int i, j, indx;
	double minabs;
	complex double Lambda[6], Evl[6][6], Evr[6][6];
	complex double Munit[6][6] = {
			{ 1., 0., 0., 0., 0., 0. },
			{ 0., 1., 0., 0., 0., 0. },
			{ 0., 0., 1., 0., 0., 0. },
			{ 0., 0., 0., 1., 0., 0. },
			{ 0., 0., 0., 0., 1., 0. },
			{ 0., 0., 0., 0., 0., 1. }
	};
	complex double KTEvr[6];
	int k;

	clapack_complex_eigensystem((complex double*)KT,(complex double *)Munit, 6,
			Lambda, (complex double *)Evl, (complex double *)Evr);

	minabs = cabs(Lambda[0]);
	for (i=0; i<6; i++) {
		if (cabs(Lambda[i]) < minabs) {
			indx = i;
			minabs = cabs(Lambda[i]);
		}
		/* test and print eigenvector */
		if (DEBUG) { // DEBUG
			for (j=0; j<6; j++) {
				KTEvr[j] = 0;
				for (k=0; k<6; k++) {
					KTEvr[j] += (KT[j][k] - Lambda[i] * Munit[j][k]) * Evr[i][k];
				}
			}
			debugmsg(1,logfile,"lambda=%11.3e%+10.3ei   Evr = ",Lambda[i]);
			for (j=0; j<6; j++) {
				debugmsg(1,logfile,"%12.3e%+11.3ei ",creal(Evr[i][j]),cimag(Evr[i][j]));
			}
			debugmsg(1,logfile,"  (KT - l * 1) * ev = ");
			for (j=0; j<6; j++) {
				debugmsg(1,logfile,"%12.3e%+11.3ei ",creal(KTEvr[j]),cimag(KTEvr[j]));
			}
			debugmsg(1,logfile,"\n");
		}
	}
	debugmsg(DEBUG,logfile,"\n");

	/* set displacement at upper and lower surface. reverse sign on lower surface (see rokhlin/wang) */
	for (i=0; i<3; i++) {
		displ_u[i] = Evr[indx][i];
		displ_l[i] = -Evr[indx][i+3];
	}
}

/* calculate displacement vector on upper (displ_u) and lower (displ_l)
 * laminate surface from
 *
 *     KT * | displ_u | = | sigma_u | = | 0 |
 *          | displ_l |   | sigma_l |   | 0 |
 *
 * The determinant of KT is 0, that means KT is singular */
void wang_surface_displacement2(
		complex double KT[6][6], int u_preset, int r_delete,
		complex double displ_u[3],
		complex double displ_l[3]) {
	int i, ioff;
	complex double LHS[5][5], LHSI[5][5], Rhs[5], u5[5], displ[6], sigma[6], det;

	/* delete the column corresponding to u_x from KT
	 * set right hand side of the equation to the column
	 * of KT, which corresponds to u_x */
	reduce_singular_soe((complex double *)KT, 6, r_delete, u_preset,
			(complex double *)LHS, Rhs);

	/* solve system of equations. */
	clapack_complex_inverse((complex double*)LHS,(complex double *)LHSI,5);
	complex_vecmul((complex double *)LHSI,Rhs,u5,5);

	/* compose complete polarisation vectors for upper
	 * and lower surface of the laminate */
	ioff = 0;
	for (i=0; i<3; i++) {
		if (i != u_preset) {
			displ_u[i] = u5[i-ioff];
		} else {
			displ_u[i] = 1.;
			ioff = 1;
		}
	}
	for (i=0; i<3; i++) {
		if ((i+3) != u_preset) {
			displ_l[i] = u5[i+3-ioff];
		} else {
			displ_l[i] = 1.;
			ioff = 1;
		}
	}

	/* KT * displ must be zero */
	if (DEBUG) { // DEBUG
		for (i=0; i<3; i++) {
			displ[i]   = displ_u[i];
			displ[i+3] = displ_l[i];
		}
		det = clapack_complex_det((complex double *)KT,6);
		complex_vecmul((complex double *)KT,displ,sigma,6);
		debugmsg(1,logfile,"det=%10.3e%+11.3ei - ",det);
		for (i=0; i<6; i++) {
			debugmsg(1,logfile,"%10.3e%+11.3ei ",creal(sigma[i]),cimag(sigma[i]));
		}
		debugmsg(1,logfile,"\n");
	}
}
